# -*-python-*-

__package__    = "pocketmod/__init__.py"
__version__    = "0.2"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/pocketmod"
__date__       = "$Date: 2008/11/24 05:39:41 $"
__copyright__  = "Copyright (c) 2007-2008 Aaron Straup Cope. BSD license : http://www.modestmaps.com/license.txt"

import Image
import ImageDraw
import cairo
from reportlab.pdfgen import canvas

import math
import array

class pocketmod :

    def __init__ (self, paper_width, paper_height, paper_margin=0, paper_dpi=144) :

        self.paper_height = paper_height
        self.paper_width = paper_width
        self.paper_margin = paper_margin
        self.paper_dpi = paper_dpi

        # note the flip
        
        self.width = int(self.paper_height * self.paper_dpi)
        self.height = int(self.paper_width * self.paper_dpi)        
        self.margin = int(self.paper_margin * self.paper_dpi)

        self.sect_width = int((self.width - (self.margin * 4)) / 4)
        self.sect_height = int((self.height - (self.margin * 2)) / 2)

        self.draw_page_numbers = True
        
        self.draw_instructions = True
        self.instructions_method = 'usletter'
        
        # used to map a page number "index"
        # to a section key
        
        # this assumes somthing like:
        # 
        # a | b
        # c | d
        # e | f
        # g | h
        
        self.page_map = {
            '1':'c',
            '2':'e',
            '3':'g',
            '4':'h',
            '5':'f',
            '6':'d',
            '7':'b',
            '8':'a',
        }

        # the sect map maps section keys (a,b,c)
        # to hashes of meta data for that section
        # stuff like height, width and offsets

        self.sect_map = {
            'a' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 0},
            'b' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 0},
            
            'c' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : self.sect_width},
            'd' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : self.sect_width}, 
            
            'e' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 2 * self.sect_width},
            'f' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 2 * self.sect_width},
            
            'g' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 3 * self.sect_width},
            'h' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 3 * self.sect_width}
            }

        # pieces is a list of list mapping
        # rows and cols to rows
        #
        # maps a given piece/page to it's
        # row and col index

        self.pages = []
        self.toc_pages = []

        self.sheets = []        
        self.sheets_rendered = []
        
        self.pieces = []
        self.pieces_map = {}
        
        self.count_pages = 0
        self.count_sheets = 0
        
    # ##########################################################
    
    def layout (self) :

        count_pages = len(self.pages)
        count_toc_pages = len(self.toc_pages)
        count_total = count_pages + count_toc_pages

        # print "count pages %s" % count_pages
        # print "count_toc_pages %s" % count_toc_pages
        
        # can has blank book...
        
        if count_total == 0 :
            count_total = 8
            
        # print "%s + %s = %s" % (count_pages, count_toc_pages, count_total)
        
        count_sheets = int(math.ceil(float(count_total) / float(8)))

        # print "count sheets: %s" % count_sheets

        total_pages = count_sheets * 8
        last_sect = total_pages - count_toc_pages

        # print "last sect %s" % last_sect
        
        first_page = 1
        last_page = total_pages + 1
        offset = 4

        # print "first page: %s" % first_page
        # print "last page: %s" % last_page
        
        for current_sheet in range(1, (count_sheets + 1)) :

            pm = Image.new('RGBA', (self.height, self.width), 'white')

            for i in range(first_page, (first_page + offset)) :

                page_idx = i - 1
                
                if i == 1 :
                    self.draw_title_page()

                elif page_idx > (count_pages - 1) :
                    self.draw_blank_page(pm, i)
                    
                else :
                    self.draw_page(pm, i)
        
            for i in range((last_page - offset), last_page) :

                page_idx = i - 1

                if i == total_pages and self.draw_instructions:
                    self.draw_instructions_page(pm, i)
                    
                elif i < total_pages and i >= last_sect :
                    idx = (total_pages - i) - 1
                    self.draw_toc_page(pm, i, idx)

                elif page_idx > (count_pages - 1) :
                    self.draw_blank_page(pm, i)
                    
                else :
                    self.draw_page(pm, i)

            self.sheets_rendered.append(pm)
    
            first_page += offset
            last_page -= offset

    # ##########################################################
    
    def save (self, out) :

        # reportlab docs say out can be any IO
        # thingy or a path
        
        # note the part where we (re) flip x, y
        # and height, width; see notes in setup()
        
        pdf = canvas.Canvas(out, pagesize=(self.height, self.width))
        
        for im in self.sheets_rendered :    
            pdf.drawInlineImage(im, 0, 0)
            pdf.showPage()
    
        pdf.save()

    # ##########################################################
    
    def draw_title_page (self) :
        pass

    # ##########################################################
    
    def draw_instructions_page (self, pm, i) :

        try :
            from instructions import diagram
        except Exception, e:
            print e
            return
        
        page_idx = str(self.idx_for_page_number(i))
        
        sect = self.page_map[page_idx]
        meta = self.sect_map[sect]

        w = self.sect_width - (self.margin * 2)
        h = self.sect_height - (self.margin * 2)
        
        howto = diagram(w, h)
        howto.method = self.instructions_method
        
        img = howto.draw()

        if not img :
            return

        img = self.rotate_image(img, meta['rotate'])
        (w,h) = img.size
        
        x = int(meta['x'])
        y = int(meta['y'])
        
        margin_x = int((self.sect_height - w) / 2)
        margin_y = int((self.sect_width - h) / 2)        

        x += margin_x
        y += margin_y
        
        pm.paste(img, (x, y))

    # ##########################################################

    def draw_toc_page (self, pm, current_page, idx) :

        page_idx = str(self.idx_for_page_number(current_page))
        
        sect = self.page_map[page_idx]
        meta = self.sect_map[sect]

        img = self.toc_pages[idx]

        if self.draw_page_numbers :
            img = self.draw_pg_number(img, current_page)
            
        img = self.rotate_image(img, meta['rotate'])
    
        x = int(meta['x'])
        y = int(meta['y'])

        x += self.margin
        y += self.margin

        if sect in ('e', 'f', 'g', 'h') :
            y += (self.margin * 2)
        
        pm.paste(img, (x,y))

    # ##########################################################
    
    def draw_page (self, pm, current_page) :

        page_idx = str(self.idx_for_page_number(current_page))
        
        sect = self.page_map[page_idx]
        meta = self.sect_map[sect]
    
        img = self.pages[ (current_page - 1) ]

        if self.draw_page_numbers:
            img = self.draw_pg_number(img, current_page)

        img = self.rotate_image(img, meta['rotate'])
    
        x = int(meta['x'])
        y = int(meta['y'])

        x += self.margin
        y += self.margin

        if sect in ('e', 'f', 'g', 'h') :
            y += (self.margin * 2)
        
        pm.paste(img, (x,y))

    # ##########################################################
    
    def draw_blank_page (self, pm, current_page) :

        page_idx = str(self.idx_for_page_number(current_page))
        
        sect = self.page_map[page_idx]
        meta = self.sect_map[sect]

        img = Image.new('RGBA', (self.sect_width, self.sect_height), 'white')

        if self.draw_page_numbers :
            img = self.draw_pg_number(img, current_page)
            
        img = self.rotate_image(img, meta['rotate'])
    
        x = int(meta['x'])
        y = int(meta['y'])

        x += self.margin
        y += self.margin

        if sect in ('e', 'f', 'g', 'h') :
            y += (self.margin * 2)
        
        pm.paste(img, (x,y))

    # ##########################################################

    def draw_pg_number (self, im, pg_number) :

        (w, h) = im.size

        if pg_number % 2 :
            # odd
            x = w - 30
        else :
            # even
            x = 30
        
        y = h - 30
        
        pm = self.pil2cairo(im)
    
        ctx = cairo.Context(pm)        
        ctx.move_to(x, y)
        ctx.rectangle(x - 10, y - 10, 20, 20)
        ctx.set_source_rgb(55, 55, 55)
        ctx.fill()
        
        ctx = cairo.Context(pm)
        ctx.rectangle(x - 10, y - 10, 20, 20)
        ctx.set_source_rgb(0, 0, 0)
        ctx.set_line_width(2)
        ctx.stroke()
        
        ctx = cairo.Context(pm)        
        ctx.set_source_rgb(0, 0, 0)    
        
	offset_x, offset_y = self.offsets_for_page_id(ctx, pg_number)
                
        x = x + offset_x
        y = y - offset_y

        ctx.move_to(x, y)
        ctx.show_text(unicode(pg_number))

        return self.cairo2pil(pm)

    # ##########################################################
    
    def rotate_image (self, img, degrees) :
        return img.transpose(degrees)

    # ##########################################################
    
    def idx_for_page_number(self, pg) :
    
        while pg > 8 :
            pg -= 8

        return pg

    # ##########################################################

    def pil2cairo(self, im) :
        
        (w,h) = im.size
    
	mode = cairo.FORMAT_ARGB32

        data = im.tostring()
        a = array.array('B', data)

        return cairo.ImageSurface.create_for_data (a, mode, w, h, (w * 4))

    # ##########################################################
    
    def cairo2pil(self, surface) :

	mode='RGBA'
        
        width = surface.get_width()
        height = surface.get_height()
    
        return Image.frombuffer(mode, (width, height), surface.get_data(), "raw", mode, 0, 1)

    # ##########################################################
    
    def offsets_for_page_id(self, ctx, pg_number) :

        # http://www.tortall.net/mu/wiki/CairoTutorial#tips-and-tricks
    
        fascent, fdescent, fheight, fxadvance, fyadvance = ctx.font_extents()
        
        xbearing, ybearing, width, height, xadvance, yadvance = (
            ctx.text_extents(unicode(pg_number)))
        
        x = 0.5 - xbearing - width / 2
        y = fdescent + fheight / 2

        # wtf do I need to adjust these again by hand?
        
    	return ((x - 1), (y - 10))

    # ##########################################################
    
if __name__ == '__main__' :

    pm = pocketmod(8.5, 11, 0)
    pm.layout()
    pm.save("/home/asc/Desktop/blank.pdf")
